// This file is part of the PiDuce project (http://www.cs.unibo.it/Piduce/)
// developed at the Department of Computer Science of Bologna.
// Authors: Samuele Carpineti, Cosimo Laneve, Leonardo Mezzina, Luca Padovani
//
// See Copyright for the status of this software.

using System;
using BoPi.Common;

namespace BoPi.Types
{
  class EmptyLabel : ILabelsSet
  {
    public ILabelsSet Intersect(ILabelsSet label)
    { return this; }

    public bool IsEmpty()
    { return true; }

    public bool Includes(ILabelsSet labels)
    { return false; }
  }

  class VoidLabel : ILabelsSet 
  {
    public ILabelsSet Intersect(ILabelsSet label)
    {
      if (label is VoidLabel)
        return this;
      else return new EmptyLabel();
    }
    public bool IsEmpty()
    { return false; }

    public bool Includes(ILabelsSet labels)
    { return labels is VoidLabel; }		
  }
  class IntLiteralLabel : ILabelsSet
  {
    private int val;

    public IntLiteralLabel(int val)
    { this.val=val; }

    public ILabelsSet Intersect(ILabelsSet label)
    {
      if ((label is IntLiteralLabel) && 
        (((IntLiteralLabel)label).val==val))
        return label;
      else if (label is IntLabel) return this;
      else return new EmptyLabel();			
    }

    public bool IsEmpty()
    { return false; }

    public bool Includes(ILabelsSet labels)
    {
      if (labels is IntLiteralLabel) return ((IntLiteralLabel)labels).val == val;
      else return false;
    }
  }

  class IntLabel : ILabelsSet
  {
    public ILabelsSet Intersect(ILabelsSet label)
    {
      if (label is IntLabel) return this;
      else if (label is IntLiteralLabel) return label;
      else return new EmptyLabel();
    }
    
    public bool IsEmpty()
    { return false; }

    public bool Includes(ILabelsSet labels)
    {
      if (labels is IntLabel || labels is IntLiteralLabel) return true;
      else return false;
    }
  }


  class StringLiteralLabel : ILabelsSet
  {
    private String val;
    
    public StringLiteralLabel(String val)
    { this.val=val; }

    public ILabelsSet Intersect(ILabelsSet label)
    {
      if ((label is StringLiteralLabel) && 
        (((StringLiteralLabel)label).val==val))
        return label;
      else if (label is StringLabel) return this;
      else return new EmptyLabel();			
    }
    public bool IsEmpty()
    { return false; }

    public bool Includes(ILabelsSet labels)
    {
      if (labels is StringLiteralLabel) return ((StringLiteralLabel)labels).val == val;
      else return false;
    }
  }

  class StringLabel : ILabelsSet
  {
    public ILabelsSet Intersect(ILabelsSet label)
    {
      if (label is StringLabel) return this;
      else if (label is StringLiteralLabel) return label;
      else return new EmptyLabel();
    }
    public bool IsEmpty()
    { return false; }

    public bool Includes(ILabelsSet labels)
    {
      if (labels is StringLabel || labels is StringLiteralLabel) return true;
      else return false;
    }
  }

  class ChanLabel : ILabelsSet
  {
    protected Chan.CAPABILITY capability;

    public ChanLabel(Chan.CAPABILITY capability)
    { this.capability=capability; }

    public ILabelsSet Intersect(ILabelsSet label)
    {
      if (label is ChanLabel)
      {
        ChanLabel chanLabel = (ChanLabel) label; 
        if (capability == Chan.CAPABILITY.IN && chanLabel.capability != Chan.CAPABILITY.OUT) return this;
        else if (capability == Chan.CAPABILITY.OUT && chanLabel.capability != Chan.CAPABILITY.IN) return this;
        else return new ChanLabel(Chan.CAPABILITY.INOUT);
      }
      else return new EmptyLabel();
    }

    public bool IsEmpty()
    { return false; }

    public bool Includes(ILabelsSet labels)
    {
      if (labels is ChanLabel) 
        return Chan.HasMoreEqCapability(((ChanLabel)labels).capability, capability);
      else return false;
    }
  }
}
