// This file is part of the PiDuce project (http://www.cs.unibo.it/Piduce/)
// developed at the Department of Computer Science of Bologna.
// Authors: Samuele Carpineti, Cosimo Laneve, Leonardo Mezzina, Luca Padovani
//
// See Copyright for the status of this software.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using BoPi.Common;

namespace BoPi.Types
{ 
  public class SubtypingChecker
  {
    private EmptynessChecker ec = new EmptynessChecker();
    private DeterminednessChecker dc = new DeterminednessChecker();
    private ICollection<Assumption> aF = new List<Assumption>();
    private ICollection<Assumption> aT = new List<Assumption>();
    public SubtypingChecker(EmptynessChecker ec, DeterminednessChecker dc)
    {
      this.ec = ec;
      this.dc = dc;
    }
    
    public SubtypingChecker(): this(new EmptynessChecker(), new DeterminednessChecker())
    {}
    
    public SubtypingChecker(EmptynessChecker ec): this(ec, new DeterminednessChecker())
    {}
    
    public SubtypingChecker(DeterminednessChecker dc): this(new EmptynessChecker(), dc)
    {}

    public bool IsSubtype(IType s, IType t)
    {
      bool res = IsSubtype(s, t, aT, aF, ec, dc);
      return res;
    }

    private static bool IsSubtype(IType s, IType t, ICollection<Assumption> aT,
      ICollection<Assumption> aF, EmptynessChecker ec, DeterminednessChecker dc)
    {       
      if (ec.IsEmpty(s)) return true;
      if (s.IsVoid() && t.IsVoid()) return true;
      else if (s.IsVoid() && (t.IsBaseType() || t.IsChan() || t.IsLabelled())) return false;
      else if (s.IsErrorType() || t.IsErrorType()) return true;
      else if (s.IsBaseType() && t.IsBaseType()) 
      {
        if (s.IsInt() && t.IsInt()) return true;
        else if (s.IsIntLiteral() && t.IsInt()) return true;
        else if (s.IsIntLiteral() && t.IsIntLiteral()) 
          return (s.AsIntLiteral().Val == t.AsIntLiteral().Val);	
        else if (s.IsString() && t.IsString()) return true;
        else if (s.IsStringLiteral() && t.IsString()) return true;
        else if (s.IsStringLiteral() && t.IsStringLiteral()) 
          return s.AsStringLiteral().Val.Equals(t.AsStringLiteral().Val);
        else return false;
      }
      else if (s.IsLabelled() && t.IsLabelled()) 
      { 
        Labelled l1 = s.AsLabelled();
        Labelled l2 = t.AsLabelled();
        return l2.Labels.Includes(l1.Labels) && IsSubtype(l1.Content, l2.Content, aT, aF, ec, dc);
      }
      else if (s.IsLabelled() && (t.IsBaseType() || t.IsChan()))
        return false;
      else if (s.IsLabelled())
        return IsSubtype(new Sequence(s.AsLabelled(), new Void()), t, aT, aF, ec, dc);
      else if (t.IsLabelled())
        return IsSubtype(s, new Sequence(t.AsLabelled(), new Void()), aT, aF, ec, dc);
      else if (s.IsBaseType() && (t.IsLabelled() || t.IsChan()))
        return false;
      else if (s.IsChan() && (t.IsLabelled() || t.IsBaseType()))
        return false;
        /* TypesAlgorithms for channels works as follow &ltS1&gt^K1 &lt: &ltS2&gt^K2 :
        * if K2 is included in or is equal to K1 then 
        * if K2={Input Capability} then S1&lt:S2 (covariance)
        * if K2={Output Capability} then S2&lt:S1 (contravariance)
        * if K2={Output Capability, InputCapability} then S2&lt:S1 &&  S1&lt:S2 (invariance)
        * else return false; */
      else if (s.IsChan() && t.IsChan()) 
      {
        Chan chan1 = s.AsChan();
        Chan chan2 = t.AsChan();
        if (Chan.HasMoreEqCapability(chan1.Capability, chan2.Capability))
        {	  
          bool isSubtype = true;
          if (chan2.HasInputCapability())
            isSubtype &= IsSubtype(chan1.Carried, chan2.Carried, aT, aF, ec, dc);
          if (chan2.HasOutputCapability())
            isSubtype &= IsSubtype(chan2.Carried, chan1.Carried, aT, aF, ec, dc);
          return isSubtype;
        }
        return false;
      }
      else 
      { 
        Assumption ass = new Assumption(s, t);
        if (aT.Contains(ass))
          return true;
        else if (aF.Contains(ass))
          return false;
        else
        {
          bool res = false;
          aT.Add(ass);
          if (s.IsSequence() && (t.IsSequence() || t.IsUnion() || t.IsConstantTypeName()))
            res = AuxSubtype(s.AsSequence(), t, aT, aF, ec, dc);
          else if (s.IsSequence()) res = false;
          else if (s.IsConstantTypeName())
            res = IsSubtype(s.AsConstantTypeName().Entry.Type, t, aT, aF, ec, dc);
          else if (s.IsUnion())
            res = IsSubtype(s.AsUnion().Fst, t, aT, aF, ec, dc) && 
              IsSubtype(s.AsUnion().Snd, t, aT, aF, ec, dc);
          else if (s.IsVoid())
          {
            if (t.IsUnion())
              res = IsSubtype(s, t.AsUnion().Fst, aT, aF, ec, dc) ||
                IsSubtype(s, t.AsUnion().Snd, aT, aF, ec, dc);
            else if (t.IsSequence())
              res = IsSubtype(s, t.AsSequence().Top, aT, aF, ec, dc) &&
                IsSubtype(s, t.AsSequence().Tail, aT, aF, ec, dc);
            else if (t.IsConstantTypeName())
              res = IsSubtype(s, t.AsConstantTypeName().Entry.Type, aT, aF, ec, dc);
          }
          else if (s.IsBaseType() || s.IsChan())
          {
            if (t.IsUnion())
              res = IsSubtype(s, t.AsUnion().Fst, aT, aF, ec, dc) ||
                IsSubtype(s, t.AsUnion().Snd, aT, aF, ec, dc);
            else if (t.IsConstantTypeName())
              res = IsSubtype(s, t.AsConstantTypeName().Entry.Type, aT, aF, ec, dc);
          }
          else if (s.IsVoid() && t.IsSequence())
            res = IsSubtype(s, t.AsSequence().Top, aT, aF, ec, dc) &&
              IsSubtype(s, t.AsSequence().Tail, aT, aF, ec, dc);
          else if (t.IsUnion())
            res = IsSubtype(s, t.AsUnion().Fst, aT, aF, ec, dc) ||
              IsSubtype(s, t.AsUnion().Snd, aT, aF, ec, dc);
          else if (t.IsConstantTypeName())
            res = IsSubtype(s, t.AsConstantTypeName().Entry.Type, aT, aF, ec, dc);
          else throw new ApplicationException("IsSubtype: Unhandled case " + s.GetType() + "?<:?" + t.GetType());
          if (!res)
          {
            aT.Remove(ass);
            aF.Add(ass);
          }
          return res;
        }
      }
    }
    /// <summary>
    /// Checks the subtyping relation between a sequence and a type.
    /// </summary>
    /// <param name="s">the sequence</param>
    /// <param name="t">the type</param>
    /// <param name="ta">a set containing typing assumptions </param>
    /// <returns>true if <c>s</c> is a subtype of <c>t</c>, false otherwise</returns>
    private static bool AuxSubtype(Sequence s, IType t, 
      ICollection<Assumption> aT, ICollection<Assumption> aF, 
      EmptynessChecker ec, DeterminednessChecker dc)
    {
      if (t.IsConstantTypeName())
        return IsSubtype(s, t.AsConstantTypeName().Entry.Type, aT, aF, ec, dc);
      else if (t.IsSequence())
        return IsSubtype(s.Top, t.AsSequence().Top, aT, aF, ec, dc) &&
          IsSubtype(s.Tail, t.AsSequence().Tail, aT, aF, ec, dc);
      else if (t.IsUnion())
      {
        if (IsValueType(s, new List<IType>()))
          return IsSubtype(s, t.AsUnion().Fst, aT, aF, ec, dc) || IsSubtype(s, t.AsUnion().Snd, aT, aF, ec, dc);
        //this simplified version of the algorithm can be applied only if t is determined 
        else if (dc.IsDetermined(t))
        {
          return IsSubtype(s, t.AsUnion().Fst, aT, aF, ec, dc) || 
            IsSubtype(s, t.AsUnion().Snd, aT, aF, ec, dc);
        }
        else if (dc.IsDetermined(t))
          return AuxDetSubTyping(s, t.AsUnion(), aT, aF, ec, dc);
        else //if (!dc.IsDetermined(t))
          return AuxNDetSubtyping(s, t.AsUnion(), aT, aF, ec, dc);
      }
      else Debug.Assert(false);
      return false;
    }

    private static bool AuxDetSubTyping(Sequence s, Union u, 
      ICollection<Assumption> aT, ICollection<Assumption> aF,
      EmptynessChecker ec, DeterminednessChecker dc) 
    {
      Labelled top = s.Top.AsLabelled();
      LabelsSet l = top.Labels;
      IList<ILabelsSet> first = dc.GetFirst(u.Fst);
      LabelsSet intersect = LabelsSet.Intersection(l, first);
      LabelsSet Difference = LabelsSet.Difference(l, first);
      bool isSubtype = true;
      Sequence seq1 = new Sequence(new Labelled(intersect, top.Content), s.Tail);
      Sequence seq2 = new Sequence(new Labelled(Difference, top.Content), s.Tail);
      if (!intersect.IsEmpty())
        isSubtype &= IsSubtype(seq1, u.Fst, aT, aF, ec, dc);
      if (!Difference.IsEmpty())
        isSubtype &= IsSubtype(seq2, u.Snd, aT, aF, ec, dc);
      return isSubtype;
    }

    private static bool AuxNDetSubtyping(Sequence s, Union u, 
      ICollection<Assumption> aT, ICollection<Assumption> aF,
      EmptynessChecker ec, DeterminednessChecker dc)
    {
      IList<Sequence> expansion = TypesOperations.Expand(u);
      expansion = Filter(s, expansion);
      IList<LabelsSet> first = new List<LabelsSet>();
      foreach (Sequence seq in expansion) first.Add(seq.Top.Labels);//TODO dc.GetFirst
      //normalize the type with respect to "first". 
      //A label of exptop is either included or disjoint from each label of first
      IType normalizedType = Normalize(s.Top.AsLabelled(), s.Tail, first);
      if (normalizedType.IsUnion())
      {
        IType tmp = normalizedType.AsUnion().Fst;
        while (normalizedType.IsUnion())
        {
          tmp = new Union(tmp, normalizedType.AsUnion().Snd);
          normalizedType = normalizedType.AsUnion().Snd;
        }
        return IsSubtype(tmp, u, aT, aF, ec, dc);
      }
      else
      {
        Debug.Assert(normalizedType.IsSequence());
        ulong subsetNum = ((ulong)1) << expansion.Count;
        bool isSub = true;
        //only for debugging purposes
        if (expansion.Count > 10)
          Console.WriteLine("Expensive subtyping: Things to prove = {0}", subsetNum);
        for (ulong c = 0; c < subsetNum && isSub; c++)
        {
          IList<IType> inter1 = new List<IType>();
          IList<IType> inter2 = new List<IType>();
          for (int i = 0; i < expansion.Count; i++)
          {
            Sequence seq = (Sequence)expansion[i];
            ulong bitmask = ((ulong)1) << i;
            if ((c & bitmask) != 0)
              inter1.Add(normalizedType.AsSequence().Top.Content);
            else inter2.Add(seq.Tail);
          }
          isSub = false;
          if (inter1.Count != 0)
          {
            IType t1 = TypesOperations.Unify(inter1);
            isSub |= IsSubtype(normalizedType.AsSequence().Top.Content, t1, aT, aF, ec, dc);
          }
          if (!isSub && inter2.Count != 0)
          {
            IType t2 = TypesOperations.Unify(inter2);
            isSub = IsSubtype(s.Tail, t2, aT, aF, ec, dc);
          }
        }
        return isSub;
      }
    }
    /// <summary>
    /// Given a labelled type <c>L[S]</c> and a set of first 
    /// <c>{L0, L1, ...., Ln}</c> splits <c>L[S]</c> is an equivalent union of
    /// schemas <c>L0'[S] + L1'[S] + ... + Ln[S]</c> such that the Intersection 
    /// between <c>Li'</c> and <c>Li</c> is either empty or included in Li' 
    /// and the union of <c>Li</c> is equal to <c>L</c>.  
    /// </summary>
    /// <param name="s">the labelled type <c>L[S]</c></param>
    /// <param name="first">a set containing AbstractLabelsSet 
    /// <c>{L0, L1, ...., Ln}</c></param>
    /// <returns>a type <c>L0'[S] + L1'[S] + ... + Ln[S]</c> such that 
    /// Intersection between <c>Li'</c> and <c>Li</c> is either empty or included
    /// in Li' and the union of <c>Li</c> is equal to <c>L</c>.
    /// </returns>
    private static IType Normalize(Labelled s, IType tail, IList<LabelsSet> first)
    {
      //The algorithm works as follows:
      //the set "labels" contains the set of labels to normalize. Initially it
      //contains the labels of "s". Then, for each label "Li" in "first"
      //each label "L" in labels it is rewritten as "L\Li" and "L^Li" (^ is the
      //Intersection). 
      IList<LabelsSet> labels = new List<LabelsSet>();
      labels.Add(s.Labels);
      for (int j = 0; j < first.Count; j++)
      {
        IList<LabelsSet> normalizedLabels = new List<LabelsSet>();
        LabelsSet fl = first[j];
        for (int i = 0; i < labels.Count; i++)
        {
          LabelsSet ls = labels[i];
          LabelsSet diff = ls.Subtract(fl);
          if (!diff.IsEmpty()) normalizedLabels.Add(diff);
          ILabelsSet inter = ls.Intersect(fl);
          if (!inter.IsEmpty()) normalizedLabels.Add((LabelsSet)inter);
        }
        labels = normalizedLabels;
      }
      //A new schema is built
      IList<IType> types = new List<IType>(labels.Count);
      for (int i=0; i < labels.Count; i++)
        types.Add(new Sequence(new Labelled(labels[i], s.Content), tail));
      return TypesOperations.Unify(types);
    }
    /// <summary>
    /// Returns true if T is generated from the following grammar:
    /// S ::= a[S],S | Int | String | <S>^K | U | n | s
    /// </summary>
    private static bool IsValueType(IType t, ICollection<IType> visited)
    {
      if (t.IsUnion()) return false;
      else if (t.IsConstantTypeName())
      {
        if (visited.Contains(t)) return true;
        else
        {
          visited.Add(t);
          return IsValueType(t.AsConstantTypeName().Entry.Type, visited);
        }
      }
      else if (t.IsSequence()) 
        return IsValueType(t.AsSequence().Top, visited) && IsValueType(t.AsSequence().Tail, visited);
      else if (t.IsLabelled())
      {
        return (t.AsLabelled().Labels is UnionLabel) && 
          (((UnionLabel)t.AsLabelled().Labels).labels.Count == 1) &&
          IsValueType(t.AsLabelled().Content, visited);
      }
      else return true;
    }
 
    private static IList<Sequence> Filter(Sequence s, IList<Sequence> types)
    {
      IList<Sequence> exp= new List<Sequence>();
      foreach (Sequence seq in types)
        if (!s.Top.Labels.Intersect(seq.Top.Labels).IsEmpty())
          exp.Add(seq);
      return exp;
    }
  }
}