// This file is part of the PiDuce project (http://www.cs.unibo.it/Piduce/)
// developed at the Department of Computer Science of Bologna.
// Authors: Samuele Carpineti, Cosimo Laneve, Leonardo Mezzina, Luca Padovani
//
// See Copyright for the status of this software.

using System;
using System.Collections.Generic;
using System.Diagnostics;
using BoPi.Common;
using System.Xml;

namespace BoPi.Types
{
	public class CompileTypeVisitor : ITypeVisitor
	{
		private XmlWriter writer;

		public CompileTypeVisitor(XmlWriter writer)
		{ this.writer = writer; }

		public void VisitBottom(Bottom t)
		{
			writer.WriteStartElement("bottom");
			writer.WriteEndElement();
		}

		public void VisitVoid(Void t)
		{
			writer.WriteStartElement("void");
			writer.WriteEndElement();
		}

		public void VisitUnion(Union t)
		{
			writer.WriteStartElement("choice");
			t.Fst.Accept(this);
			t.Snd.Accept(this);
			writer.WriteEndElement();
		}

		public void VisitConstantTypeName(ConstantTypeName t)
    {
			Debug.Assert(t.Entry.Used);
      writer.WriteStartElement("schemaref");
      writer.WriteAttributeString("name", t.Entry.FreshName);
      writer.WriteEndElement();
    }

		public void VisitSequence(Sequence t)
    {
      writer.WriteStartElement("sequence");
			t.Top.Accept(this);
			t.Tail.Accept(this);
      writer.WriteEndElement();
    }

		public void VisitLabelled(Labelled t)
    {
      writer.WriteStartElement("labelled");
      t.Labels.GetCode(writer);
			t.Content.Accept(this);
      writer.WriteEndElement();
    }

		public void VisitIntLiteral(IntLiteral t)
    {
      writer.WriteStartElement("intLit");
      writer.WriteString(t.Val.ToString());
      writer.WriteEndElement();
    }

		public void VisitIntType(IntType t)
    {
      writer.WriteStartElement("int");
      writer.WriteEndElement();
    }

		public void VisitStringLiteral(StringLiteral t)
    {
      writer.WriteStartElement("strLit");
      writer.WriteCData(t.Val);
      writer.WriteEndElement();
    }

		public void VisitStringType(StringType t)
    {
      writer.WriteStartElement("string");
      writer.WriteEndElement();
    }

		public void VisitChan(Chan t)
    {
      string cap = "?";
      if (t.Capability == Chan.CAPABILITY.OUT) cap = "O";
      if (t.Capability == Chan.CAPABILITY.INOUT) cap = "IO";
      if (t.Capability == Chan.CAPABILITY.IN) cap = "I";
      writer.WriteStartElement("chan");
      writer.WriteAttributeString("capability", cap);
      t.Carried.Accept(this);
      writer.WriteEndElement();
    }

		public void VisitErrorType(ErrorType t)
    {
      writer.WriteStartElement("error");
      writer.WriteEndElement();
    }

		public void VisitBind(Bind t)
		{
			// LUCA: todo: use entry instead of index?
			writer.WriteStartElement("bind");
     	writer.WriteAttributeString("index", t.Index.ToString());
			t.Content.Accept(this);
			writer.WriteEndElement();
		}
	}

	public class ResolveNamesTypeVisitor : ITypeVisitor
	{
		private TypeSymbolTable st;

		public ResolveNamesTypeVisitor(TypeSymbolTable st)
		{ this.st = st; }

		public void VisitBind(Bind t)
		{ t.Content.Accept(this); }

		public void VisitBottom(Bottom t)
		{ }

		public void VisitUnion(Union t)
		{
			t.Fst.Accept(this);
			t.Snd.Accept(this);
		}

		public void VisitVoid(Void t)
		{ }

		public void VisitConstantTypeName(ConstantTypeName t)
		{ t.Entry = (ISymbolTableTypeEntry) st.Lookup(t.Name); }

		public void VisitSequence(Sequence t)
		{
			t.Top.Accept(this);
			t.Tail.Accept(this);
		}

		public void VisitLabelled(Labelled t)
		{ t.Content.Accept(this); }

		public void VisitIntLiteral(IntLiteral t)
		{ }

		public void VisitIntType(IntType t)
		{ }

		public void VisitStringLiteral(StringLiteral t)
		{ }

		public void VisitStringType(StringType t)
		{ }

		public void VisitChan(Chan t)
		{ t.Carried.Accept(this); }

		public void VisitErrorType(ErrorType t)
		{ }
	}

	public class ToStringTypeVisitor : ITypeVisitor
	{
		private System.Text.StringBuilder builder = new System.Text.StringBuilder();

		public override string ToString()
		{ return builder.ToString(); }

		public void VisitBottom(Bottom t)
		{ builder.Append("empty"); }

		public void VisitVoid(Void t)
		{ builder.Append("void"); }

		public void VisitUnion(Union t)
		{
			t.Fst.Accept(this);
			builder.Append("+");
			t.Snd.Accept(this);
		}

		public void VisitConstantTypeName(ConstantTypeName t)
		{ builder.Append(t.Name); }

		public void VisitSequence(Sequence t)
		{ 
			t.Top.Accept(this);
			builder.Append(",");
			t.Tail.Accept(this);
		}

		public void VisitLabelled(Labelled t)
    { 
			builder.Append("{" + t.Labels + "}[");
			t.Content.Accept(this);
			builder.Append("]");
		}

		public void VisitIntLiteral(IntLiteral t)
		{ builder.Append(t.Val.ToString()); }

		public void VisitIntType(IntType t)
		{ builder.Append("int"); }

		public void VisitStringLiteral(StringLiteral t)
		{ builder.Append("\"" + t.Val + "\""); }

		public void VisitStringType(StringType t)
		{ builder.Append("string"); }

		public void VisitChan(Chan t)
		{ 
      builder.Append("<");
			t.Carried.Accept(this);
			builder.Append(">");
			switch (t.Capability) {
      case Chan.CAPABILITY.IN:
        builder.Append("^I");
				break;
      case Chan.CAPABILITY.OUT:
        builder.Append("^O");
				break;
      case Chan.CAPABILITY.INOUT:
        builder.Append("^IO");
				break;
			}
		}

		public void VisitErrorType(ErrorType t)
		{ builder.Append("error"); }

		public void VisitBind(Bind t)
		{
			builder.Append(t.Index.ToString() + ":");
			t.Content.Accept(this);
		}
	}

	public class StripTypeVisitor : ITypeVisitor
	{
		private IType result = null;

		public IType Result { get { return result; } }

		public IType Visit(IType t)
		{
			t.Accept(this);
			return result;
		}

		public void VisitBottom(Bottom t)
		{ result = t; }

		public void VisitVoid(Void t)
		{ result = t; }

		public void VisitUnion(Union t)
		{
			IType fst = Visit(t.Fst);
			IType snd = Visit(t.Snd);
			if (Object.ReferenceEquals(fst, t.Fst) && Object.ReferenceEquals(snd, t.Snd))
				result = t;
			else
				result = new Union(fst, snd);
		}

		public void VisitConstantTypeName(ConstantTypeName t)
		{ result = t; }

		public void VisitSequence(Sequence t)
		{
			IType top = Visit(t.Top);
			IType tail = Visit(t.Tail);
			if (Object.ReferenceEquals(top, t.Top) && Object.ReferenceEquals(tail, t.Tail))
				result = t;
			else	
				result = new Sequence(t.Top, tail);
		}

		public void VisitLabelled(Labelled t)
		{
			IType content = Visit(t.Content);
			if (Object.ReferenceEquals(content, t.Content))
				result = t;
			else
				result = new Labelled(t.Labels, content);
		}

		public void VisitIntLiteral(IntLiteral t)
		{ result = t; }

		public void VisitIntType(IntType t)
		{ result = t; }

		public void VisitStringLiteral(StringLiteral t)
		{ result = t; }

		public void VisitStringType(StringType t)
		{ result = t; }

		public void VisitChan(Chan t)
		{ result = t; }

		public void VisitErrorType(ErrorType t)
		{ result = t; }

		public void VisitBind(Bind t)
		{ t.Content.Accept(this); }
	}

#if false
	public class StripTypeVisitor : ITypeVisitor
	{
		private Stack<IType> res = new Stack<IType>();

		public IType Result { get { return res.Pop(); } }

		public void VisitBottom(Bottom t)
		{ res.Push(t); }

		public void VisitVoid(Void t)
		{ res.Push(t); }

		public void VisitUnion(Union t)
		{
			t.Fst.Accept(this);
			t.Snd.Accept(this);
			IType snd = res.Pop();
			IType fst = res.Pop();
			if (Object.ReferenceEquals(fst, t.Fst) && Object.ReferenceEquals(snd, t.Snd))
				res.Push(t);
			else
				res.Push(new Union(fst, snd));
		}

		public void VisitConstantTypeName(ConstantTypeName t)
		{ res.Push(t); }

		public void VisitSequence(Sequence t)
		{
			t.Top.Accept(this);
			t.Tail.Accept(this);
			IType tail = res.Pop();
			IType top = res.Pop();
			if (Object.ReferenceEquals(top, t.Top) && Object.ReferenceEquals(tail, t.Tail))
				res.Push(t);
			else	
				res.Push(new Sequence(top.AsLabelled(), tail));
		}

		public void VisitLabelled(Labelled t)
		{
			t.Content.Accept(this);
			IType content = res.Pop();
			if (Object.ReferenceEquals(content, t.Content))
				res.Push(t);
			else
				res.Push(new Labelled(t.Labels, content));
		}

		public void VisitIntLiteral(IntLiteral t)
		{ res.Push(t); }

		public void VisitIntType(IntType t)
		{ res.Push(t); }

		public void VisitStringLiteral(StringLiteral t)
		{ res.Push(t); }

		public void VisitStringType(StringType t)
		{ res.Push(t); }

		public void VisitChan(Chan t)
		{ res.Push(t); }

		public void VisitErrorType(ErrorType t)
		{ res.Push(t); }

		public void VisitBind(Bind t)
		{ t.Content.Accept(this); }
	}
#endif

  public abstract class Type: IType 
  {	
    public virtual IntType AsInt()
    { 
      throw new ApplicationException(this + " is not Int");
    }
    public virtual StringType AsString()
    { 
      throw new ApplicationException(this + " is not String");
    }
    public virtual IntLiteral AsIntLiteral()
    { 
      throw new ApplicationException(this + " is not IntLiteral");
    }
    public virtual StringLiteral AsStringLiteral()
    { 
      throw new ApplicationException(this + " is not StringLiteral");
    }
    public virtual Labelled AsLabelled()
    { 
      throw new ApplicationException(this + " is not Labelled");
    }
    public virtual Sequence AsSequence()
    { 
      throw new ApplicationException(this + " is not Sequence");
    }
    public virtual Union AsUnion()
    { 
      throw new ApplicationException(this + " is not Union");
    }
    public virtual ConstantTypeName AsConstantTypeName()
    { 
      throw new ApplicationException(this + " is not ConstantTypeName");
    }
    public virtual Chan AsChan()
    { 
      throw new ApplicationException(this + " is not Chan");
    }
    public virtual Void AsVoid()
    { 
      throw new ApplicationException(this + " is not Void");
    }
    public virtual Bottom AsBottom()
    { 
      throw new ApplicationException(this + " is not Bottom");
    }
    public virtual bool IsInt()
    { 
      return false;
    }
    public virtual bool IsString()
    { 
      return false;
    }
    public virtual bool IsIntLiteral()
    { 
      return false;
    }
    public virtual bool IsStringLiteral()
    { 
      return false;
    }
    public virtual bool IsLabelled()
    { 
      return false;
    }
    public virtual bool IsSequence()
    { 
      return false;
    }
    public virtual bool IsUnion()
    { 
      return false;
    }
    public virtual bool IsConstantTypeName()
    { 
      return false;
    }
    public virtual bool IsBaseType()
    { 
      return false;
    }
    public virtual bool IsChan()
    { 
      return false;
    }
    public virtual bool IsVoid()
    { 
      return false;
    }
    public virtual bool IsBottom()
    { 
      return false;
    }
    public virtual bool IsErrorType()
    { 
      return false;
    }

    public override bool Equals(Object obj)
    {
      if (obj is IType)
        return TypesOperations.Equals(this, (IType) obj);
      else
        return false;
    }

    public override int GetHashCode()
    { return TypesOperations.GetHashCode(this); }

		public IType Strip()
		{
			StripTypeVisitor visitor = new StripTypeVisitor();
			Accept(visitor);
			IType res = visitor.Result;
			return res;
		}

    public override String ToString()
		{
			ToStringTypeVisitor visitor = new ToStringTypeVisitor();
			Accept(visitor);
			return visitor.ToString();
		}

    public void Compile(XmlWriter writer)
		{
			CompileTypeVisitor visitor = new CompileTypeVisitor(writer);
			Accept(visitor);
		}

		public abstract void Accept(ITypeVisitor visitor);

		public void ResolveNames(TypeSymbolTable st)
		{
			ResolveNamesTypeVisitor visitor = new ResolveNamesTypeVisitor(st);
			Accept(visitor);
		}
  }
}
