// This file is part of the PiDuce project (http://www.cs.unibo.it/Piduce/)
// developed at the Department of Computer Science of Bologna.
// Authors: Samuele Carpineti, Cosimo Laneve, Leonardo Mezzina, Luca Padovani
//
// See Copyright for the status of this software.

using System;
using System.Diagnostics;
using BoPi.Common;

namespace BoPi.Types
{
	public class Bind : Type
	{
		private readonly int index;
		private readonly IType content;

		public int Index { get { return index; } }
		public IType Content { get { return content; } }

		public Bind(int index, IType content)
		{
			this.index = index;
			this.content = content;
		}

		public override void Accept(ITypeVisitor visitor)
		{ visitor.VisitBind(this); }
	}

  /// <summary>
  /// Class that represent the bottom type
  /// </summary>
  public class Bottom: Type 
  {
    public override bool IsBottom()
    { return true; }

    public override Bottom AsBottom()
    { return this; }

		public override void Accept(ITypeVisitor visitor)
		{ visitor.VisitBottom(this); }
  }
  /// <summary>
  /// Class that represent the union type
  /// </summary>
  public class Union: Type 
  {
    private IType fst;
    private IType snd;

    public Union(IType fst, IType snd) 
    {
      this.fst = fst;
      this.snd = snd;
    }
    public IType Fst { get { return fst; } }
    public IType Snd { get { return snd; } }

    public override bool IsUnion()
    { return true; }

    public override Union AsUnion()
    { return this; }

		public override void Accept(ITypeVisitor visitor)
		{ visitor.VisitUnion(this); }
  }

  /// <summary>
  /// Denotes the empty type (it recognizes the empty string)
  /// </summary>
  public class Void : Type 
  {	
    public override bool IsVoid()
    { return true; }

    public override Void AsVoid()
    { return this; }

		public override void Accept(ITypeVisitor visitor)
		{ visitor.VisitVoid(this); }
  }
  
  public class ConstantTypeName : Type 
  {
    private readonly String name;
		private ISymbolTableTypeEntry entry;
    
		public ConstantTypeName(String name) : this(name, null)
		{ }

    public ConstantTypeName(String name, ISymbolTableTypeEntry entry) 
    {
      this.name = name;
			this.entry = entry;
    }
    
    public String Name { get { return name; } }

		public ISymbolTableTypeEntry Entry {
			get { return entry; }
			set { entry = value; }
		}

    public override bool IsConstantTypeName()
    { return true; }

    public override ConstantTypeName AsConstantTypeName()
    { return this; }

		public override void Accept(ITypeVisitor visitor)
		{ visitor.VisitConstantTypeName(this); }
  }

  /// <summary>
  /// Denotes sequences.
  /// </summary>
  public class Sequence :Type 
  {
    private Labelled top;
    private IType tail;

    public Sequence(Labelled top, IType tail) 
    {
      this.top = top;
      this.tail = tail;
    }
    public Labelled Top { get { return top; } }

    public IType Tail { get { return tail; } }

    public override bool IsSequence()
    { return true; }

    public override Sequence AsSequence()
    { return this; }

		public override void Accept(ITypeVisitor visitor)
		{ visitor.VisitSequence(this); }
  }
    
  public class Labelled: Type 
  {
    private LabelsSet matchedLabels;
    private IType content;
		
    public Labelled(LabelsSet labels, IType content) 
    {
      this.matchedLabels = labels;
      this.content = content;
    }

    public LabelsSet Labels 
    { get { return matchedLabels; } }

    public IType Content { get { return content; } }

    public override bool IsLabelled()
    { return true; }

    public override Labelled AsLabelled()
    { return this; }

		public override void Accept(ITypeVisitor visitor)
		{ visitor.VisitLabelled(this); }
  }
  
  public abstract class BaseType: Type
  {
    public override bool IsBaseType()
    { return true; }
  }

  public class IntLiteral:BaseType 
  {
    private readonly int val;
		
    public IntLiteral(int val)
    { this.val=val; }

    public int Val { get { return val; } }

    public override bool IsIntLiteral()
    { return true; }

    public override IntLiteral AsIntLiteral()
    { return this; }	

		public override void Accept(ITypeVisitor visitor)
		{ visitor.VisitIntLiteral(this); }
  }
    
  public class IntType:BaseType 
  {
    public override bool IsInt()
    { return true; }

    public override IntType AsInt()
    { return this; }

		public override void Accept(ITypeVisitor visitor)
		{ visitor.VisitIntType(this); }
  }
  
  public class StringLiteral:BaseType 
  {
    private readonly String val;

    public StringLiteral(String val)
    { this.val=val; }

    public String Val { get { return val; }}

    public override bool IsStringLiteral()
    { return true; }

    public override StringLiteral AsStringLiteral()
    { return this; }

		public override void Accept(ITypeVisitor visitor)
		{ visitor.VisitStringLiteral(this); }
  } 

  public class StringType:BaseType 
  {
    public override bool IsString()
    { return true; }

    public override StringType AsString()
    { return this; }

		public override void Accept(ITypeVisitor visitor)
		{ visitor.VisitStringType(this); }
  }
  
  /**
   * Denotes the channel type. In particular a channel can have one of the following capabilities: 
   * input capability, output capability, both. 
   */
  public class Chan:Type 
  {
    public enum CAPABILITY { OUT = +1, IN = -1, INOUT = +2 }
    private IType carried;
    private CAPABILITY capability;
		
    public Chan(IType carried, int capability)
    {
      this.carried = carried;
      if (capability == +1) this.capability = CAPABILITY.OUT;
      else if (capability == -1) this.capability = CAPABILITY.IN;
      else this.capability = CAPABILITY.INOUT;
      this.carried = carried;
    }
    public Chan(IType carried, Chan.CAPABILITY capability) 
    {
      this.carried = carried;
      this.capability = capability;
    }
    public IType Carried { get { return carried; } }
				
    public CAPABILITY Capability { get { return capability; } }

    public override bool IsChan()
    { return true; }

    public override Chan AsChan()
    { return this; }

    public bool HasInputCapability() 
    { return (capability == CAPABILITY.IN) || (capability == CAPABILITY.INOUT); }

    public bool HasOutputCapability() 
    { return (capability == CAPABILITY.OUT) || (capability == CAPABILITY.INOUT); }

    public bool HasInputOutputCapability() 
    { return (capability == CAPABILITY.INOUT); }

    public static bool HasMoreEqCapability(Chan.CAPABILITY cap1, Chan.CAPABILITY cap2) 
    { return ((Math.Abs((int)cap1) > Math.Abs((int)cap2)) || (cap1 == cap2)); }
        
		public override void Accept(ITypeVisitor visitor)
		{ visitor.VisitChan(this); }
  }

  public class ErrorType:Type 
  {		
    public override bool IsErrorType()
    { return true; }

		public override void Accept(ITypeVisitor visitor)
		{ visitor.VisitErrorType(this); }
  }
}
